% ------------------------------------------------------------------------
%   FILE:       RISK.PL
%   AUTHOR:     Tim Duncan
%   UPDATED:    Sun Aug  9 18:01:08 BST 1987
%   PURPOSE:    Risk analysis for ATTENDING, returning augmented tree.
% ------------------------------------------------------------------------


:- op(100,fx,*).
:- op(100,xfy,:).       % connects text to nodes of augmented tree
                        % and conditional risks in pmframes.



/* augment_tree/4
Mode: augment_tree(+,-,+,-).
Args: augment_tree(Node,Risk,Pat,Aug)
    Node = current node being investigated
    Risk = risk at this node
    Pat  = the tree below this node
    Aug  = the augmented tree returned (?)
Entry point, also used recursively for subtrees.
*/

augment_tree(_,_,[],[]) :- !.   % lowest magnitude risk.
augment_tree(Node,Risk,[Subnode|Others],[Subtree|Othertree]) :- !,
    node(Node,Arclist),
    head(Arc,Subnode),
    member(arc(Arc,Next,Sub_ref),Arclist),
    examine_choice(Arclist,Subnode,Subrisk,Subtree),
    augment_tree(Next,Otherrisk,Others,Othertree),
    highest_risk(Subrisk,Otherrisk,Risk).     % Highest risk over AND-ed nodes




/* augment_node(+,+,+,+,-,-)
    If no better choices then return choice + prose.
    If choice has no risk (ie: none, lowb, modb) then alternatives must
        be better (ie: we're not interested in equivalents).
    Else, return choice + list of best alternatives + prose
*/

augment_node(Choice,C_risk,[],Best_risk,Choice,C_risk) :- !.
augment_node(Choice,C_risk,Better,Best_risk,Choice,C_risk) :-
    less_or_equal_risk(C_risk,none),
    less_or_equal_risk(lowb,Best_risk).         % not good enough
augment_node(Choice,C_risk,Better,Best_risk,
    ['ALTERNATIVES',[Choice|Better]],Best_risk).




/* examine_choice/4
Mode: examine_choice(+,+,-,-).
Args: examine_choice(Arcs,Taken,Risk,Return).
    Arcs = arcs available at this point
    Taken = subtree representing choices taken
    Risk = overall risk of this choice
    Return = augmented form of subtree
The first clause deals with terminal nodes. Otherwise the second clause
looks first at the subplan.

Note: the tree originally uses node-names for reference. The augmented
tree uses the action routine ref instead as this is unique (and prosenet
requires it). The Refs are introduced in the list of betters produced by
get_better_choices/4, and in the args given to augment_node/6
*/

examine_choice(Arclist,[Choice],Risk,Return) :- !,
    delete(arc(Choice,To,Ref),Arclist,Alternatives),
    evaluate_node_risk(Ref,C_risk,Prose),
    get_better_choices(Alternatives,C_risk,Better,Best_risk),
    cprule_comparison(Ref,Better,B_list),
    augment_node([Ref:C_risk:Prose],C_risk,B_list,Best_risk,Return,Risk).
examine_choice(Arclist,[Daughter|Tree],Risk,Return) :-
    augment_tree(Daughter,Drisk,Tree,Dreturn),          % inherited risk
    delete(arc(Daughter,To,Ref),Arclist,Alternatives),
    evaluate_node_risk(Ref,Own_risk,Prose),               % own risk
    highest_risk(Own_risk,Drisk,Trisk),
    get_better_choices(Alternatives,Trisk,Better,B_risk),
    cprule_comparison(Ref,Better,B),
    augment_node([Ref:Trisk:Prose|Dreturn],Trisk,B,B_risk,Return,Risk).





/* evaluate_node_risk/3
    Collect the set of risks (since there may be risks and benefits), and
    the overall risk of this node (ie. the highest). The third arg returned
    is the set of Risk:Description pairs
*/

evaluate_node_risk(Node_ref,Risk,Set) :-
    fast_setof(R:P,lookup_node_risk(Node_ref,R,P),Set),
    high_risk_list(Set,_,Risk:Prose), !.
evaluate_node_risk(Node_ref,none,[]).                % default: no risk



/* lookup_node_risk/3
    Called by evaluate_node_risk/3 to obtain the set of risks.
*/

lookup_node_risk(Node_ref,extremer,[]) :-
    recorded(block,blocked(Node_ref),_), !.                % arc unavailable (cut)
lookup_node_risk(Node_ref,Risk,Prose) :-
    recorded(risk,risk(Node_ref,Risk,Prose),_).            % check database
lookup_node_risk(Node_ref,Risk,Prose) :-
    recorded(risk,risk(Node_ref:Cond,Risk,Prose),_),       % conditional risk
    check_context(Cond).



/* high_risk_list/3
Mode: high_risk_list(+,?,-)
Args: high_risk_list(Set,Acc,Highest)
    Set = List of Risk:Prose pairs for comparison
    Acc = highest Risk:Prose pair found so far (variable at start)
    Highest = return highest at end
(Note: since ':' is xfy, then Prose may in fact be Val:Prose)
*/

high_risk_list([],Acc,Acc).
high_risk_list([R:P|Rest],AR:AP,HR:HP) :-
    less_or_equal_risk(AR,R), !,      % pass this down as highest risk
    high_risk_list(Rest,R:P,HR:HP).
high_risk_list([R:P|Rest],AR:AP,HR:HP) :- !,
    high_risk_list(Rest,AR:AP,HR:HP).  % do not change accumulator



/* get_better_choices/4
Mode: get_better_choices(+,+,-,-)
Args: get_better_choices(Arcs,Comp,Better,Risk)
    Arcs = alternative arcs that could have been taken
    Comp = risk value to compare alternatives against
    Better = list of better alternatives
    Risk = lowest risk found
Note: in the returned structure, the Ref replaces the Node-name
*/

get_better_choices([],Risk,[],Risk).
get_better_choices([arc(Choice,_,Ref)|Others],Risk,Return,Best_risk) :-
    get_better_choices(Others,Risk,Better,Brisk),
    evaluate_node_risk(Ref,Crisk,Prose),
    compare_risks(Crisk,Brisk,[Ref:Crisk:Prose],Better,Return,Best_risk).





/* compare_risks/6
Mode: compare_risks(+,+, +,+, -,-)
Args: compare_risks(Nr,Ar, N,A, R,Rr)
Compare the node (N) with the list collected so far (A) using
their respective risk values (Nr and Ar). If they are equivalent
then the node is added to the list. If the node has a lower risk
then it replaces the accumulated list, else it is discarded. As
risks are added, the CP rules are checked to see if any of those
accumulated can be discarded.
*/

compare_risks(Same,Same,Ch,Bh,[Ch|Bh],Same) :- !,         % equivalent - add ?
    eliminate_by_cprule(Ch,Bh,Return).
compare_risks(Crisk,Brisk,Choice,_,[Choice],Crisk) :-
    less_or_equal_risk(Crisk,Brisk), !.                   % better - replace
compare_risks(_,Brisk,_,Better,Better,Brisk).             % worse - ignore





/* eliminate_by_cprule(Candidate,Others,Return)
   If one technique is preferred over another in a given disease context,
   then the less-preferred technique is dropped. Note: the Ref is used now
   rather than the node-name.
*/

eliminate_by_cprule([Rf:Ps],Bh,Return) :-
    cprule(Rf,Less,Cond,Why),            % Eliminate earlier choice
    member([Less:P],Bh),
    check_context(Cond),
    delete([Less:P],Bh,New_Bh), !,
    eliminate_by_cprule([Rf:Ps],New_Bh,Return).
eliminate_by_cprule([Rf:Ps],Bh,Bh) :-
    cprule(Pref,Rf,Cond,Why),            % Previous technique preferred
    member([Pref:P],Bh),
    check_context(Cond), !.
eliminate_by_cprule(Candidate,Others,[Candidate|Others]).   % add to list






/* cprule_comparison(Chosen,Alts,Return)
   If there is a CPRULE which explains why one of the alternatives is better
   than the chosen one, then add this explanation to the prose. "Chosen" is
   the technique chosen by the medic and so is not deleted.
*/

cprule_comparison(Ref,[],[]).
cprule_comparison(Ref,[[Pref:R:Ps]|Rest],[[Pref:R:['CP'(Cond:Why)|Ps]]|More]) :-
    cprule(Pref,Ref,Cond,Why),   % Say why PREF is a better technique
    check_context(Cond), !,
    cprule_comparison(Ref,Rest,More).
cprule_comparison(Ref,[[A:Ps]|Rest],[[A:Ps]|More]) :- !,
    cprule_comparison(Ref,Rest,More).




/* check_context/1
Mode: check_context(+)
Args: check_context(List).
    List = conditions to check
Predicate succeeds if all the conditions are true, ie: disease is present,
or where condition is not(X), is not present.
*/

check_context([]).
check_context([not(X)|Rest]) :- !,
    not(recorded(patient,X,_)), !,
    check_context(Rest).
check_context([X|Rest]) :-
    recorded(patient,X,_), !,
    check_context(Rest).




/* ---------------------- RISK UTILITIES ------------------------- */



/* highest_risk/3
Mode: highest_risk(+,?,-)
Args: highest_risk(A,B,C)
Compare two risks and return the highest. The second argument
is sometimes called uninstantiated, eg: the first time by
augment_tree, since it represents the accumulated highest risk
over 'ANDed' nodes.
*/                    


highest_risk(A,B,B) :-
    not(var(B)),
    less_or_equal_risk(A,B).
highest_risk(A,B,A).



/* This predicate encodes the partial order used for risks */

less_or_equal_risk(X,X) :- !.
less_or_equal_risk(A,B) :-
    before(B,A,[extremer,highr,modr,lowr,none,lowb,modb,highb]), !.




risky(X) :-
    less_or_equal_risk(lowr,X).

benefit(X) :-
    less_or_equal_risk(X,lowb).



/* sortrisk/2
    Sort risks/benefits into order: extremer, highr, modr, lowr, none,
    lowb, modb. Essentially, this is qsort/3 from Warren's thesis, but
    adapted to compare risks. Note that also, the risks have prose attatched.
*/


sortrisk(U,S) :- qsort(U,S,[]).

qsort([X|L],R,R0) :-
        partition(L,X,L1,L2),
        qsort(L2,R1,R0),
        qsort(L1,R,[X|R1]).
qsort([],R,R).

partition([X|L],Y,[X|L1],L2) :-
        ordered(X,Y), !,
        partition(L,Y,L1,L2).
partition([X|L],Y,L1,[X|L2]) :-
        partition(L,Y,L1,L2).
partition([],_,[],[]).


/* Note: this ordering ensures that for two equal risks,
if one has a CP rule attatched, then it will come later
ie: less risk. Also note that this predicate has to cope
with the risk to compare being first item when called by
btlist/4, and second when called by organise_minors */


ordered(T1:R:XP,T2:R:YP) :- !,    % format given by
    not(member('CP'(_),XP)).      % organise_minors
ordered(T1:X:XP,T2:Y:YP) :- !,
    less_or_equal_risk(Y,X).

ordered(R:XP,R:YP) :- !,          % format given by
    not(member('CP'(_),XP)).      % btlist/4
ordered(X:XP,Y:YP) :-
    less_or_equal_risk(Y,X).
